#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>


struct BitmapFileHeader {
	uint8_t sig[2];
	uint32_t fileSz;
	uint8_t rfu[4];
	uint32_t dataOfst;
	uint32_t headerSz;			//40
	int32_t width;
	int32_t height;
	uint16_t colorplanes;		//must be one
	uint16_t bpp;
	uint32_t compression;
	uint32_t dataLen;			//may be 0
	uint32_t pixelsPerMeterX;
	uint32_t pixelsPerMeterY;
	uint32_t numColors;			//if zero, assume 2^bpp
	uint32_t numImportantColors;
	
} __attribute__((packed));

struct BitmapClutEntry {
	uint8_t b, g, r, x;
} __attribute__((packed));

enum EinkClut {
	EinkClutTwoBlacks = 0,
	EinkClutFourBlacks,
	EinkClutEightBlacks,
	EinkClutSevenBlacksAndYellow,
	EinkClutSevenBlacksAndRed,
	EinkClutSixteenBlacks,
};

void getBytes(void* dstP, unsigned len)
{
	char *dst = (char*)dstP;
	
	while (len--) {
		
		int ch;
		
		if ((ch = getchar()) == EOF) {
			fprintf(stderr, "input file ended prematurely\n");
			exit(-1);
		}
		
		*dst++ = ch;
	}
}

void putBytes(const void* srcP, unsigned len)
{
	const char *src = (char*)srcP;
	
	while (len--)
		putchar(*src++);
}

static void usage(const char *self)
{
	fprintf(stderr, "USAGE: %s [-D] <CLUT TYPE> < in.bmp > out.img\n"
		"  Input bitmap should be a 24bpp color bitmap\n"
		"  -D requests dithering\n"
		"  Valid CLUT types:\n"
		"    \"1bpp\" - Black & White\n"
		"    \"2bpp\" - 2-bit greyscale (4 blacks)\n"
		"    \"3bpp\" - 3-bit greyscale (8 blacks)\n"
		"    \"3bppY\" - 3-bit greyscale with Yellow (7 blacks plus one yellow color)\n"
		"    \"3bppR\" - 3-bit greyscale with Red (7 blacks plus one red color)\n"
		"    \"4bpp\" - 4-bit greyscale (16 blacks)\n"
		" This program does no dithering. Do that on your own time! RED should be\n"
		" #FF0000, Yellow should be #FFFF00\n"
		, self);
	
	exit(-1);
}

int main(int argc, char **argv)
{
	uint32_t c, rowBytesOut, rowBytesIn, outBpp, i, numRows;
	uint32_t numGrays, extraColor = 0;
	struct BitmapFileHeader hdr;
	const char *self = argv[0];
	enum EinkClut clutType;
	uint8_t clut[256][3];
	bool dither = false;
	
	if (argc != 2 && argc != 3)
		usage(self);
	
	if (argc == 3) {
		if (strcmp(argv[1], "-D"))
			usage(self);
		dither = true;
		argv++;
	}
	if (!strcmp(argv[1], "1bpp"))
		clutType = EinkClutTwoBlacks;
	else if (!strcmp(argv[1], "2bpp"))
		clutType = EinkClutFourBlacks;
	else if (!strcmp(argv[1], "3bpp"))
		clutType = EinkClutEightBlacks;
	else if (!strcmp(argv[1], "3bppY"))
		clutType = EinkClutSevenBlacksAndYellow;
	else if (!strcmp(argv[1], "3bppR"))
		clutType = EinkClutSevenBlacksAndRed;
	else if (!strcmp(argv[1], "4bpp"))
		clutType = EinkClutSixteenBlacks;
	else
		usage(self);
	
	getBytes(&hdr, sizeof(hdr));
	
	if (hdr.sig[0] != 'B' || hdr.sig[1] != 'M' || hdr.headerSz < 40 || hdr.colorplanes != 1 || hdr.bpp != 24 || hdr.compression) {
		fprintf(stderr, "BITMAP HEADER INVALID\n");
		return -1;
	}
	
	switch (clutType) {
		case EinkClutTwoBlacks:
			numGrays = 2;
			outBpp = 1;
			break;
		
		case EinkClutFourBlacks:
			numGrays = 4;
			outBpp = 2;
			break;
		
		case EinkClutEightBlacks:
			numGrays = 8;
			outBpp = 3;
			break;
		
		case EinkClutSevenBlacksAndYellow:
			numGrays = 7;
			extraColor = 0xffff00;
			outBpp = 3;
			break;
		
		case EinkClutSevenBlacksAndRed:
			numGrays = 7;
			extraColor = 0xff0000;
			outBpp = 3;
			break;
		
		case EinkClutSixteenBlacks:
			numGrays = 16;
			outBpp = 4;
			break;
	}
	
	
	rowBytesIn = (hdr.width * hdr.bpp + 31) / 32 * 4;
	rowBytesOut = (hdr.width * outBpp + 31) / 32 * 4;
	
	numRows = hdr.height < 0 ? -hdr.height : hdr.height;
	hdr.bpp = outBpp;
	hdr.numColors = 1 << outBpp;
	hdr.numImportantColors = 1 << outBpp;
	hdr.dataOfst = sizeof(struct BitmapFileHeader) + 4 * hdr.numColors;
	
	putBytes(&hdr, sizeof(hdr));
	
	//emit & record grey clut entries
	for (i = 0; i < numGrays; i++) {
		
		uint32_t val = 255 * i / (numGrays - 1);
		
		putchar(val);
		putchar(val);
		putchar(val);
		putchar(val);
		
		clut[i][0] = val;
		clut[i][1] = val;
		clut[i][2] = val;
	}
	
	//if there is a color CLUT entry, emit that
	if (extraColor) {
		
		putchar((extraColor >> 0) & 0xff);	//B
		putchar((extraColor >> 8) & 0xff);	//G
		putchar((extraColor >> 16) & 0xff);	//R
		putchar(0x00);						//A
		
		clut[i][0] = (extraColor >> 0) & 0xff;
		clut[i][1] = (extraColor >> 8) & 0xff;
		clut[i][2] = (extraColor >> 16) & 0xff;
	}
	
	while (numRows--) {
		
		uint32_t valSoFar = 0, bytesIn = 0, bytesOut = 0, bitsSoFar = 0;
		
		for (c = 0; c < hdr.width; c++, bytesIn += 3) {
			int64_t bestDist = 0x7fffffffffffffffll;
			uint8_t rgb[3], bestIdx = 0;
			int32_t ditherFudge = 0;
			
			getBytes(rgb, sizeof(rgb));
			
			if (dither)
				ditherFudge = (rand() % 255 - 127) / (int)numGrays;
			
			for (i = 0; i < (1 << outBpp); i++) {
				int64_t dist = 0;
				
				dist += (rgb[0] - clut[i][0] + ditherFudge) * (rgb[0] - clut[i][0] + ditherFudge) * 4750ll;
				dist += (rgb[1] - clut[i][1] + ditherFudge) * (rgb[1] - clut[i][1] + ditherFudge) * 47055ll;
				dist += (rgb[2] - clut[i][2] + ditherFudge) * (rgb[2] - clut[i][2] + ditherFudge) * 13988ll;
				
				if (dist < bestDist) {
					bestDist = dist;
					bestIdx = i;
				}
			}
			
			valSoFar = (valSoFar << outBpp) | bestIdx;
			bitsSoFar += outBpp;
			
			if (bitsSoFar >= 8) {
				putchar(valSoFar >> (bitsSoFar -= 8));
				valSoFar &= (1 << bitsSoFar) - 1;
				bytesOut++;
			}
		}
		if (bitsSoFar) {
			putchar(valSoFar);
			bytesOut++;
		}
		
		while (bytesIn++ < rowBytesIn)	
			getchar();
		while (bytesOut++ < rowBytesOut)
			putchar(0);
	}
	fprintf(stderr, "SUCCESS\n");
	return 0;
}